@@ -1,11 +1,10 @@
 from __future__ import annotations
 import copy
 import re
-from typing import Union, Callable, Optional, List
-from collections import OrderedDict
+from typing import Union, Callable, Optional
+from collections import OrderedDict, Counter
 
 import numpy as np
-import pandas as pd
 import elementy
 
 
@@ -24,7 +23,7 @@ class Alloy:
 
     """
 
-    class Composition(dict):
+    class Composition(OrderedDict):
         """Atomic percentages of elements in an alloy.
 
         :group: alloy
@@ -41,7 +40,9 @@ class Alloy:
             super().__init__(value, *args, **kwargs)
             self.on_change = on_change
 
-        def __setitem__(self, element: str, percentage: float):
+        def __setitem__(
+            self, element: str, percentage: float, respond_to_change=True
+        ):
             """
             Set the percentage value of an element in the composition.
             Remove elements from the composition if below minimum threshold.
@@ -51,11 +52,13 @@ class Alloy:
             else:
                 if element in self.keys():
                     super().__delitem__(element)
-            self.on_change()
+            if hasattr(self, "on_change") and respond_to_change:
+                self.on_change()
 
-        def __delitem__(self, element):
+        def __delitem__(self, element, respond_to_change=True):
             super().__delitem__(element)
-            self.on_change()
+            if hasattr(self, "on_change") and respond_to_change:
+                self.on_change()
 
     def __init__(
         self,
@@ -92,7 +95,12 @@ class Alloy:
         elif isinstance(other, (str, dict)):
             return self.to_string() == Alloy(other).to_string()
         else:
-            raise NotImplementedError
+            return False
+
+    def __add__(self, other):
+        from . import generate
+
+        return generate.mixture([self, other])
 
     def __hash__(self):
         return hash(self.to_string())
@@ -208,11 +216,16 @@ class Alloy:
         :group: alloy.utils
         """
 
+        if len(self.composition) < 1:
+            return
+
         self.rescaling = True
 
         constraints_applied = False
         while not self.constraints_satisfied():
 
+            self.determine_percentage_constraints()
+
             constraints_applied = True
 
             precedence = {}
@@ -251,14 +264,7 @@ class Alloy:
                         inclusive=True,
                     ):
                         continue
-                    # print(
-                    #     element,
-                    #     self.composition[element],
-                    #     "out of",
-                    #     direction,
-                    #     "range",
-                    #     self.constraints["local_percentages"][element],
-                    # )
+
                     for other_element in reverse_precedence_order:
                         if (
                             element != other_element
@@ -273,27 +279,6 @@ class Alloy:
                             ):
                                 continue
 
-                            # if (
-                            #     other_element
-                            #     in self.constraints["local_percentages"]
-                            # ):
-                            #     print(
-                            #         other_element,
-                            #         self.composition[other_element],
-                            #         "in",
-                            #         direction,
-                            #         "range",
-                            #         self.constraints["local_percentages"][
-                            #             other_element
-                            #         ],
-                            #     )
-                            # else:
-                            #     print(
-                            #         other_element,
-                            #         self.composition[other_element],
-                            #         "not constrained",
-                            #     )
-
                             if direction == "min":
                                 donate_percentage(
                                     self.composition,
@@ -325,10 +310,19 @@ class Alloy:
 
         if not constraints_applied:
             self.clamp_composition()
-            self.round_composition()
+        self.round_composition()
+
+        self.reorder_composition()
 
         delattr(self, "rescaling")
 
+    def reorder_composition(self):
+        ordered_elements = sorted(
+            self.composition, key=self.composition.get, reverse=True
+        )
+        for element in ordered_elements:
+            self.composition.move_to_end(element)
+
     def constrain_max_elements(self):
         """Removes elements from an alloy if there are more elements than
         allowed by the max_elements constraint."""
@@ -390,8 +384,10 @@ class Alloy:
                 - multiple_round(
                     self.total_percentage,
                     self.constraints["percentage_step"],
+                    self.constraints["digits"],
                 )
             )
+
             if discrepancy > self.constraints["percentage_step"]:
                 # print("Sum != 1", self.total_percentage, discrepancy)
                 # print(self.composition)
@@ -449,8 +445,23 @@ class Alloy:
                                         self.composition[element]
                                         > self.composition[superior_element]
                                     ):
+                                        # print(
+                                        #     "element above superior",
+                                        #     element,
+                                        #     superior_element,
+                                        #     self.composition[element],
+                                        #     self.constraints[
+                                        #         "local_percentages"
+                                        #     ][element]["max"],
+                                        #     self.composition[superior_element],
+                                        # )
+                                        # print(self.composition)
+                                        # print(self.constraints)
+                                        # print()
                                         satisfied = False
                                         break
+                                if not satisfied:
+                                    break
 
                     elif (
                         self.constraints["local_percentages"][element]["min"]
@@ -488,30 +499,91 @@ class Alloy:
         :group: alloy.utils
         """
 
-        sigfigs = 3
+        digits = 4
+        percentage_step = 0.01
         if self.constraints is not None:
-            if "sigfigs" in self.constraints:
-                sigfigs = self.constraints["sigfigs"]
+            if "digits" in self.constraints:
+                digits = self.constraints["digits"]
+            if "percentage_step" in self.constraints:
+                percentage_step = self.constraints["percentage_step"]
+
+        while (
+            np.abs(
+                1
+                - multiple_round(
+                    self.total_percentage, percentage_step, digits
+                )
+            )
+            >= percentage_step
+        ):
 
-        while np.abs(1 - self.total_percentage) > 10 ** (-sigfigs):
             current_total = self.total_percentage
+            elements = self.elements
 
-            tmp_composition = copy.deepcopy(self.composition)
-            for element in self.composition:
+            clamped_rounded_total = 0
+            for element in elements:
                 clamped_value = self.composition[element] / current_total
 
                 if self.constraints is not None:
                     if element in self.constraints["percentages"]:
-                        tmp_composition[element] = max(
+                        self.composition[element] = max(
                             clamped_value,
                             self.constraints["percentages"][element]["min"],
                         )
                     else:
-                        tmp_composition[element] = clamped_value
+                        self.composition[element] = clamped_value
                 else:
-                    tmp_composition[element] = clamped_value
+                    self.composition[element] = clamped_value
 
-            self.composition = tmp_composition
+                if element in self.composition:
+                    clamped_rounded_total += multiple_round(
+                        self.composition[element], percentage_step, digits
+                    )
+
+            while (
+                round(np.abs(1 - clamped_rounded_total), digits)
+                >= percentage_step
+            ):
+                elements = self.elements
+                for element in elements:
+                    if element not in self.composition:
+                        continue
+
+                    if self.constraints is not None:
+                        if element in self.constraints["percentages"]:
+                            change = max(
+                                [
+                                    percentage_step,
+                                    max(
+                                        [
+                                            0,
+                                            self.constraints["percentages"][
+                                                element
+                                            ]["max"]
+                                            - self.composition[element],
+                                        ]
+                                    ),
+                                ]
+                            )
+                        else:
+                            change = percentage_step
+                    else:
+                        change = percentage_step
+
+                    change = min([change, np.abs(1 - clamped_rounded_total)])
+                    if clamped_rounded_total > 1:
+                        change *= -1
+
+                    change = multiple_round(change, percentage_step, digits)
+
+                    self.composition.__setitem__(
+                        element,
+                        self.composition[element] + change,
+                        respond_to_change=False,
+                    )
+                    clamped_rounded_total += change
+                    if np.abs(1 - clamped_rounded_total) > percentage_step / 2:
+                        break
 
     def determine_percentage_constraints(self):
         """Determine the current constraints on an alloy composition.
@@ -536,18 +608,18 @@ class Alloy:
 
         for element in self.constraints["percentages"]:
             if element in self.composition:
+                element_constraint = self.constraints["percentages"][element]
                 if (
-                    self.constraints["percentages"][element]["precedence"]
+                    element_constraint["precedence"]
                     <= lowest_precedence["precedence"]
-                    and self.constraints["percentages"][element]["precedence"]
-                    > 0
+                    and element_constraint["precedence"] > 0
                     and self.composition[element]
                     < lowest_precedence["percentage"]
                 ):
 
-                    lowest_precedence["precedence"] = self.constraints[
-                        "percentages"
-                    ][element]["precedence"]
+                    lowest_precedence["precedence"] = element_constraint[
+                        "precedence"
+                    ]
                     lowest_precedence["element"] = element
                     lowest_precedence["percentage"] = self.composition[element]
 
@@ -565,6 +637,7 @@ class Alloy:
                 ]:
                     if e in self.composition:
                         superior_elements.append(self.composition[e])
+
                 if len(superior_elements) > 0:
                     tmp_percentages[element]["max"] = min(
                         [
@@ -572,16 +645,49 @@ class Alloy:
                             min(superior_elements),
                         ]
                     )
-                else:
-                    tmp_percentages[element]["max"] = 0
+
+                    tmp_percentages[element]["min"] = min(
+                        [
+                            tmp_percentages[element]["min"],
+                            tmp_percentages[element]["max"],
+                        ]
+                    )
+
+            # if (
+            #     len(
+            #         self.constraints["percentages"][element][
+            #             "inferior_elements"
+            #         ]
+            #     )
+            #     > 0
+            # ):
+            #     inferior_elements = []
+            #     for e in self.constraints["percentages"][element][
+            #         "inferior_elements"
+            #     ]:
+            #         if e in self.composition:
+            #             inferior_elements.append(self.composition[e])
+
+            #     if len(inferior_elements) > 0:
+            #         tmp_percentages[element]["min"] = max(
+            #             [
+            #                 self.constraints["percentages"][element]["min"],
+            #                 max(inferior_elements),
+            #             ]
+            #         )
+
+            #         tmp_percentages[element]["max"] = max(
+            #             [
+            #                 tmp_percentages[element]["min"],
+            #                 tmp_percentages[element]["max"],
+            #             ]
+            #         )
 
         if lowest_precedence["element"] is not None:
             for element in self.composition:
                 if element not in tmp_percentages:
-                    maxPercent = lowest_precedence["percentage"]
-
                     tmp_percentages[element] = {
-                        "max": maxPercent,
+                        "max": lowest_precedence["percentage"],
                         "min": 0,
                         "precedence": 0,
                         "superior_elements": [],
@@ -599,6 +705,42 @@ class Alloy:
 
         self.constraints["local_percentages"] = tmp_percentages
 
+        if self.constraints["max_elements"] == len(
+            self.constraints["local_percentages"]
+        ):
+            total_max = 0
+            for element in self.composition:
+                if element in self.constraints["local_percentages"]:
+                    total_max += self.constraints["local_percentages"][
+                        element
+                    ]["max"]
+            while total_max < 1:
+                for element in self.constraints["local_percentages"]:
+                    if element not in self.composition:
+                        value = min(
+                            [
+                                self.constraints["local_percentages"][element][
+                                    "max"
+                                ],
+                                1 - total_max,
+                            ]
+                        )
+                        self.composition.__setitem__(
+                            element, value, respond_to_change=False
+                        )
+                        total_max += value
+
+                if total_max < 1:
+                    element_to_add = np.random.choice(
+                        self.constraints["allowed_elements"], 1
+                    )[0]
+                    if element_to_add not in self.composition:
+                        value = np.random.uniform()
+                        self.composition.__setitem__(
+                            element_to_add, value, respond_to_change=False
+                        )
+                        total_max += value
+
     def to_string(self) -> str:
         """Convert the alloy composition to a string
 
@@ -658,35 +800,58 @@ class Alloy:
         elif len(self.composition) == 1:
             self.composition[self.elements[0]] = 1.0
 
+        digits = 4
+        percentage_step = 0.0001
         if self.constraints is not None:
-            sigfigs = self.constraints["sigfigs"]
-        else:
-            sigfigs = 3
+            if "digits" in self.constraints:
+                digits = self.constraints["digits"]
+            if "percentage_step" in self.constraints:
+                percentage_step = self.constraints["percentage_step"]
+
+        needs_rounding = False
+        for element in self.elements:
+            percentage_str = str(self.composition[element])
+            length = len(percentage_str)
+            if "." in percentage_str:
+                length -= 2
+            if length > digits:
+                easy_round = False
+                easy_rounders = ["0", "9"]
+                count = Counter(percentage_str)
+                for c in easy_rounders:
+                    if c in count and not easy_round:
+                        count[c] /= len(percentage_str)
+                        if count[c] > 0.5:
+                            easy_round = True
+                            self.composition[element] = multiple_round(
+                                self.composition[element],
+                                percentage_step,
+                                digits,
+                            )
+                if not easy_round:
+                    needs_rounding = True
+                    break
+
+        if not needs_rounding:
+            return
 
         integer_parts = []
         decimal_parts = []
 
         for element in self.elements:
+            percentage = str(self.composition[element] * (10**digits))
+            split_percentage = percentage.split(".")
 
-            rounded_percentage = normal_round(
-                self.composition[element], sigfigs
-            )
+            integer_parts.append(int(split_percentage[0]))
 
-            if rounded_percentage > 0:
+            if len(split_percentage) > 1:
+                decimal_parts.append(
+                    float(percentage) - float(split_percentage[0])
+                )
+            else:
+                decimal_parts.append(0.0)
 
-                percentage = str(self.composition[element] * (10**sigfigs))
-                split_percentage = percentage.split(".")
-
-                integer_parts.append(int(split_percentage[0]))
-
-                if len(split_percentage) > 1:
-                    decimal_parts.append(
-                        float(percentage) - float(split_percentage[0])
-                    )
-                else:
-                    decimal_parts.append(0.0)
-
-        undershoot = (10**sigfigs) - sum(integer_parts)
+        undershoot = (10**digits) - sum(integer_parts)
 
         if undershoot > 0:
             if self.constraints is not None:
@@ -722,19 +887,23 @@ class Alloy:
                         filtered_precedence_order[i]
                         in self.constraints["local_percentages"]
                     ):
-                        if (integer_parts[decimal_part_index] + 1) / (
-                            10**sigfigs
-                        ) > self.constraints["local_percentages"][
-                            filtered_precedence_order[i]
-                        ][
-                            "max"
-                        ]:
-                            constraints_violated = True
+                        if decimal_part_index < len(integer_parts):
+
+                            if (integer_parts[decimal_part_index] + 1) / (
+                                10**digits
+                            ) > self.constraints["local_percentages"][
+                                filtered_precedence_order[i]
+                            ][
+                                "max"
+                            ]:
+                                constraints_violated = True
 
                 if not constraints_violated:
                     if decimal_part_index < len(integer_parts):
                         integer_parts[decimal_part_index] += 1
                         undershoot -= 1
+                else:
+                    break
                 i += 1
 
                 if i >= len(filtered_precedence_order):
@@ -743,20 +912,25 @@ class Alloy:
             i = 0
             for element in self.elements:
                 if i < len(integer_parts):
-                    self.composition[element] = normal_round(
-                        integer_parts[i] / (10**sigfigs), sigfigs
+                    self.composition[element] = multiple_round(
+                        integer_parts[i] / (10**digits),
+                        percentage_step,
+                        digits,
                     )
                 else:
                     self.composition[element] = 0
                 i += 1
 
         else:
-
             i = 0
             for element in self.elements:
 
                 if i < len(integer_parts):
-                    rounded_value = integer_parts[i] / (10**sigfigs)
+                    rounded_value = multiple_round(
+                        integer_parts[i] / (10**digits),
+                        percentage_step,
+                        digits,
+                    )
 
                     if self.constraints is not None:
                         if element in self.constraints["local_percentages"]:
@@ -775,6 +949,21 @@ class Alloy:
 
                 i += 1
 
+        deviation = 1 - self.total_percentage
+        abs_deviation = np.abs(deviation)
+        if abs_deviation >= percentage_step:
+            num_steps = int(abs_deviation // percentage_step)
+
+            j = 0
+            for _ in range(num_steps):
+                step = percentage_step
+                if deviation < 0:
+                    step *= -1
+
+                self.composition[self.elements[j]] += step
+                if j > self.num_elements:
+                    j = 0
+
 
 def parse_composition(composition: Union[str, dict, Alloy]) -> dict:
     """Parse elemental percentages of an alloy from input
@@ -795,6 +984,9 @@ def parse_composition_string(composition_string: str) -> dict:
     :group: alloy.utils
     """
 
+    if len(composition_string) == 0:
+        return None
+
     if "(" in composition_string:
 
         sub_composition_percentage = (
@@ -814,9 +1006,19 @@ def parse_composition_string(composition_string: str) -> dict:
         for element in sub_composition:
             sub_composition[element] *= sub_composition_percentage
 
-        remaining_composition_string = composition_string.split(")")[1][
-            len(str(int(sub_composition_percentage * 100))) :
-        ]
+        pre_sub_composition_string = composition_string.split(")")[0].split(
+            "("
+        )[0]
+        post_sub_composition_string = composition_string.split(")")[1].split(
+            "("
+        )[0]
+
+        remaining_composition_string = ""
+        if re.search("[a-zA-Z]", pre_sub_composition_string) is not None:
+            remaining_composition_string += pre_sub_composition_string
+        if re.search("[a-zA-Z]", post_sub_composition_string) is not None:
+            remaining_composition_string += post_sub_composition_string
+
         remaining_composition = parse_composition_string_block(
             remaining_composition_string
         )
@@ -868,19 +1070,20 @@ def parse_composition_dict(composition: dict) -> dict:
     :group: alloy.utils
     """
 
-    tmp_composition = copy.deepcopy(composition)
+    if len(composition) == 0:
+        return None
 
     needRescale = False
-    for element in tmp_composition:
-        if tmp_composition[element] > 1:
+    for element in composition:
+        if composition[element] > 1:
             needRescale = True
             break
 
     if needRescale:
-        for element in tmp_composition:
-            tmp_composition[element] /= 100.0
+        for element in composition:
+            composition[element] /= 100.0
 
-    return filter_order_composition(tmp_composition)
+    return filter_order_composition(composition)
 
 
 def filter_order_composition(composition: dict) -> OrderedDict:
@@ -913,7 +1116,6 @@ def parse_constraints(
     percentages: dict = {},
     allowed_elements: list = [e for e in elementy.PeriodicTable().elements],
     disallowed_elements: list = [],
-    sigfigs: int = 3,
     percentage_step: float = 0.01,
 ) -> dict:
     """Parse constraint rules from input
@@ -932,7 +1134,7 @@ def parse_constraints(
             List of elements allowed in an alloy
         disallowed_elements
             List of elements not allowed in an alloy
-        sigfigs
+        digits
             Number of significant figures to consider for percentages
         percentage_step
             Increment between percentages
@@ -962,6 +1164,7 @@ def parse_constraints(
 
     for element in percentages:
         percentages[element]["superior_elements"] = []
+        percentages[element]["inferior_elements"] = []
         for other_element in percentages:
             if element != other_element:
                 if (
@@ -971,14 +1174,35 @@ def parse_constraints(
                     percentages[element]["superior_elements"].append(
                         other_element
                     )
+                elif (
+                    percentages[element]["precedence"]
+                    > percentages[other_element]["precedence"]
+                ):
+                    percentages[element]["inferior_elements"].append(
+                        other_element
+                    )
 
-        inferior_min = percentages[element]["min"]
-        for other_element in percentages:
-            if other_element not in percentages[element]["superior_elements"]:
-                inferior_min = max(
-                    [inferior_min, percentages[other_element]["min"]]
+        if len(percentages[element]["superior_elements"]) > 0:
+            inferior_min = percentages[element]["min"]
+            for other_element in percentages:
+                if (
+                    other_element
+                    not in percentages[element]["superior_elements"]
+                ):
+                    inferior_min = max(
+                        [inferior_min, percentages[other_element]["min"]]
+                    )
+            if inferior_min < percentages[element]["max"]:
+                percentages[element]["min"] = inferior_min
+
+        if len(percentages[element]["inferior_elements"]) > 0:
+            superior_min = percentages[element]["min"]
+            for other_element in percentages[element]["inferior_elements"]:
+                superior_min = max(
+                    [superior_min, percentages[other_element]["min"]]
                 )
-        percentages[element]["min"] = inferior_min
+            if superior_min > percentages[element]["min"]:
+                percentages[element]["min"] = superior_min
 
     min_sum = sum([percentages[e]["min"] for e in percentages])
     if min_sum > 1:
@@ -990,13 +1214,17 @@ def parse_constraints(
         if element in allowed_elements:
             allowed_elements.remove(element)
 
+    digits = len(str(percentage_step))
+    if "." in str(percentage_step):
+        digits -= 2
+
     return {
         "percentages": percentages,
         "local_percentages": percentages,
         "allowed_elements": allowed_elements,
         "min_elements": min_elements,
         "max_elements": max_elements,
-        "sigfigs": sigfigs,
+        "digits": digits,
         "percentage_step": percentage_step,
     }
 
@@ -1042,12 +1270,6 @@ def donate_percentage(
 
     if donor in constraints["local_percentages"]:
 
-        # if (
-        #     constraints["local_percentages"][recipient]["precedence"]
-        #     < constraints["local_percentages"][donor]["precedence"]
-        # ):
-        #     return
-
         if composition[donor] > constraints["local_percentages"][donor]["max"]:
             donor_surplus = (
                 composition[donor]
@@ -1063,12 +1285,12 @@ def donate_percentage(
     else:
         donor_surplus = composition[donor]
 
-    if normal_round(donor_surplus, constraints["sigfigs"] + 1) > 0:
+    if normal_round(donor_surplus, constraints["digits"] + 1) > 0:
         if recipient_deficit <= 0 or recipient_deficit >= donor_surplus:
             donation = min([0.1, donor_surplus])
         else:
             donation = min([0.1, recipient_deficit])
-        donation = normal_round(donation, constraints["sigfigs"] + 1)
+        donation = normal_round(donation, constraints["digits"] + 1)
         # print(recipient, "gets", donation, "from", donor)
         composition[recipient] += donation
         composition[donor] -= donation
@@ -1109,34 +1331,34 @@ def element_is_in_range(
     if element not in constraints["local_percentages"]:
         return True
 
-    sigfigs = 3
-    if constraints is not None:
-        if "sigfigs" in constraints:
-            sigfigs = constraints["sigfigs"]
-
     if element in composition:
-        rounded_percentage = normal_round(composition[element], sigfigs)
+        percentage = composition[element]
     else:
-
-        rounded_percentage = 0
+        percentage = 0
 
     if direction == "min":
         if inclusive:
-            return rounded_percentage >= normal_round(
-                constraints["local_percentages"][element][direction], sigfigs
+            return (
+                percentage
+                >= constraints["local_percentages"][element][direction]
             )
         else:
-            return rounded_percentage > normal_round(
-                constraints["local_percentages"][element][direction], sigfigs
+            return (
+                percentage
+                > constraints["local_percentages"][element][direction]
             )
+
     elif direction == "max":
         if inclusive:
-            return rounded_percentage <= normal_round(
-                constraints["local_percentages"][element][direction], sigfigs
+            return (
+                percentage
+                <= constraints["local_percentages"][element][direction]
             )
+
         else:
-            return rounded_percentage < normal_round(
-                constraints["local_percentages"][element][direction], sigfigs
+            return (
+                percentage
+                < constraints["local_percentages"][element][direction]
             )
 
     return False
@@ -1160,7 +1382,7 @@ def normal_round(num: float, ndigits: int = 0) -> float:
 
 
 def multiple_round(
-    num: float, multiple: Union[float, int]
+    num: float, multiple: Union[float, int], digits=4
 ) -> Union[float, int]:
     """Round a number to the nearest multiple of another number.
 
@@ -1173,4 +1395,4 @@ def multiple_round(
     multiple
         Multiple-of to round towards
     """
-    return multiple * round(num / multiple)
+    return round(multiple * round(num / multiple), digits)
